use std::{
    fs::File,
    path::{Path, PathBuf},
};

use byte_unit::Byte;
use directories::ProjectDirs;
use log::debug;
use serde::{Deserialize, Serialize};

const QUAL: &str = "liw.fi";
const ORG: &str = "Ambient CI";
const APP: &str = env!("CARGO_PKG_NAME");

#[derive(Debug, Serialize, Deserialize)]
#[serde(deny_unknown_fields, default)]
struct StoredConfig {
    tmpdir: Option<PathBuf>,
    state: Option<PathBuf>,
    target: Option<String>,
    dput_target: Option<String>,
    run_ci: Option<PathBuf>,
    cpus: Option<usize>,
    memory: Option<Byte>,
    artifacts_max_size: Option<Byte>,
    cache_max_size: Option<Byte>,
}

impl Default for StoredConfig {
    fn default() -> Self {
        let dirs = ProjectDirs::from(QUAL, ORG, APP).expect("have home directory");
        Self {
            tmpdir: Some(PathBuf::from("/tmp")),
            state: dirs.state_dir().map(|x| x.into()),
            target: None,
            dput_target: None,
            run_ci: None,
            cpus: Some(1),
            memory: Some(Byte::parse_str("1 GB", true).unwrap()),
            artifacts_max_size: Some(Byte::parse_str("10 GB", true).unwrap()),
            cache_max_size: Some(Byte::parse_str("10 GB", true).unwrap()),
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(deny_unknown_fields)]
pub struct Config {
    tmpdir: PathBuf,
    state: PathBuf,
    target: Option<String>,
    dput_target: Option<String>,
    run_ci: Option<PathBuf>,
    cpus: usize,
    memory: Byte,
    artifacts_max_size: Byte,
    cache_max_size: Byte,
}

impl Config {
    pub fn from_yaml_file(filename: &Path) -> Result<Self, ConfigError> {
        let f = File::open(filename).map_err(|e| ConfigError::Read(filename.into(), e))?;
        let stored: StoredConfig =
            serde_yml::from_reader(f).map_err(|e| ConfigError::Yaml(filename.into(), e))?;
        let config: Self = stored.into();
        debug!(
            "Loaded configuration from file {}:\n{config:#?}",
            filename.display()
        );
        Ok(config)
    }

    pub fn tmpdir(&self) -> &Path {
        &self.tmpdir
    }

    pub fn state(&self) -> &Path {
        &self.state
    }

    pub fn target(&self) -> Option<&str> {
        self.target.as_deref()
    }

    pub fn dput_target(&self) -> Option<&str> {
        self.dput_target.as_deref()
    }

    pub fn run_ci(&self) -> Option<&Path> {
        self.run_ci.as_deref()
    }

    pub fn cpus(&self) -> usize {
        self.cpus
    }

    pub fn memory(&self) -> Byte {
        self.memory
    }

    pub fn artifacts_max_size(&self) -> u64 {
        self.artifacts_max_size.as_u64()
    }

    pub fn cache_max_size(&self) -> u64 {
        self.cache_max_size.as_u64()
    }
}

impl From<StoredConfig> for Config {
    fn from(stored: StoredConfig) -> Self {
        Self {
            tmpdir: stored.tmpdir.unwrap(),
            state: stored.state.unwrap(),
            target: stored.target,
            dput_target: stored.dput_target,
            run_ci: stored.run_ci,
            cpus: stored.cpus.unwrap(),
            memory: stored.memory.unwrap(),
            artifacts_max_size: stored.artifacts_max_size.unwrap(),
            cache_max_size: stored.cache_max_size.unwrap(),
        }
    }
}

impl Default for Config {
    fn default() -> Self {
        Self::from(StoredConfig::default())
    }
}

#[derive(Debug, thiserror::Error)]
pub enum ConfigError {
    #[error("failed to find home directory, while looking for configuration file")]
    ProjectDirs,

    #[error("failed to read configuration file {0}")]
    Read(PathBuf, #[source] std::io::Error),

    #[error("failed to parse configuration file as YAML: {0}")]
    Yaml(PathBuf, #[source] serde_yml::Error),
}
